package org.ginpink;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.RecognitionException;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.ginpink.expr.JExpr;

public class CheckInvariants {

	public CheckInvariants(String className, String methodName) {
		super();
		this.className = className;
		this.methodName = methodName;
	}

	private static final String GINPINK_INV_FILE = "ginpink_inv_file";
	private static final String CHECKED_GINPINK_INV_FILE = "checked_ginpink_inv_file";
	private static final String CLASS_NAME = "class_name";
	private static final String METHOD_NAME = "method_name";
	private static final String PPT_OMIT_PATTERN = "ppt_omit_pattern";
	private static final String PPT_SELECT_PATTERN = "ppt_select_pattern";
	private static final String DISCARDED_GINPINK_INV_FILE = "discarded_ginpink_inv_file";
	private static final String DEBUG = "debug";

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Options options = new Options();
		options.addOption(CLASS_NAME, true, "target class name");
		options.addOption(METHOD_NAME, true, "target method name");
		options.addOption(GINPINK_INV_FILE, true,
				"input ginpink invariants file name");
		options.addOption(CHECKED_GINPINK_INV_FILE, true,
				"output *checked* ginpink invariants file name");
		options.addOption(DISCARDED_GINPINK_INV_FILE, true,
				"output *discarded* ginpink invariants file name");

		options.addOption(
				PPT_OMIT_PATTERN,
				true,
				"Do not process program points whose names match the regular expression. This takes precedence over the --ppt-select-pattern argument.");
		options.addOption(
				PPT_SELECT_PATTERN,
				true,
				"Only process program points whose names match the regular expression. The --ppt-omit-pattern argument takes precedence over this argument");
		options.addOption(DEBUG, false,
				"enable debugging comments to standard output");

		CommandLineParser parser = new GnuParser();
		CommandLine commandLine;
		try {
			commandLine = parser.parse(options, args);

			List<String> dtraceFilenames = commandLine.getArgList();
			if (dtraceFilenames.size() == 0) {
				throw new ParseException(
						"Usage [options] <dtraceFilename1> <dtraceFilename2>");
			}

			if (!commandLine.hasOption(GINPINK_INV_FILE)) {
				throw new ParseException("Missing mandatory argument: "
						+ GINPINK_INV_FILE);
			}
			String invariantsFilename = commandLine
					.getOptionValue(GINPINK_INV_FILE);

			if (!commandLine.hasOption(CLASS_NAME)) {
				throw new ParseException("Missing mandatory argument: "
						+ CLASS_NAME);
			}
			String className = commandLine.getOptionValue(CLASS_NAME);

			if (!commandLine.hasOption(METHOD_NAME)) {
				throw new ParseException("Missing mandatory argument: "
						+ METHOD_NAME);
			}
			String methodName = commandLine.getOptionValue(METHOD_NAME);

			if (!commandLine.hasOption(CHECKED_GINPINK_INV_FILE)) {
				throw new ParseException("Missing mandatory argument: "
						+ CHECKED_GINPINK_INV_FILE);
			}

			String discarded_ginpink_inv_filename = null;
			if (commandLine.hasOption(DISCARDED_GINPINK_INV_FILE)) {
				discarded_ginpink_inv_filename = commandLine
						.getOptionValue(DISCARDED_GINPINK_INV_FILE);
			}

			if (commandLine.hasOption(DEBUG)) {
				GinPinkCheckerConfig.setDebugOption(true);
			}

			String checkedInvariantsFilename = commandLine
					.getOptionValue(CHECKED_GINPINK_INV_FILE);

			String ppt_omit_pattern = commandLine
					.getOptionValue(PPT_OMIT_PATTERN);

			String ppt_select_pattern = commandLine
					.getOptionValue(PPT_SELECT_PATTERN);

			CheckInvariants checkInvariants = new CheckInvariants(className,
					methodName);
			checkInvariants.checkInvariants(invariantsFilename,
					dtraceFilenames, checkedInvariantsFilename,
					discarded_ginpink_inv_filename, ppt_omit_pattern,
					ppt_select_pattern);

		} catch (ParseException e) {
			throw new RuntimeException(e);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException(e);
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (RecognitionException e) {
			throw new RuntimeException(e);
		}
	}

	private final String className;
	private final String methodName;

	public void checkInvariants(String ginpinkInvFilename,
			List<String> dtraceFilenames, String checkedGinpinkInvFilename,
			String discardedGinpinkInvFilename, String pptOmitPattern,
			String pptSelectPattern) throws ClassNotFoundException,
			NoSuchMethodException, IOException, RecognitionException {

		System.out.println("Starting ginpink invariant checker for class "
				+ className + " and method " + methodName);
		GinPinkChecker checker = new GinPinkChecker(className, methodName);
		if (pptOmitPattern != null) {
			checker.setPptOmitPattern(pptOmitPattern);
		}
		if (pptSelectPattern != null) {
			checker.setPptSelectPattern(pptSelectPattern);
		}
		Set<String> methodNames = checker.getLoopInvariantMethodNames();
		System.out.println("Found " + methodNames.size()
				+ " instrumentation loop methods");

		System.out.println("Reading ginpink invariant file "
				+ ginpinkInvFilename);
		checker.readInvariants(ginpinkInvFilename);

		HashMap<String, Integer> numberOfInvariants = new HashMap<String, Integer>();
		for (String methodName : methodNames) {
			List<Invariant> invariants = checker.getInvariants(methodName);
			numberOfInvariants.put(methodName, invariants.size());
			System.out.println(methodName + " -> " + invariants.size()
					+ " invariants");
		}

		for (String dtraceFilename : dtraceFilenames) {
			System.out.println("Reading dtrace file " + dtraceFilename);
			checker.checkInvariants(dtraceFilename);
			for (String methodName : methodNames) {
				List<Invariant> invariants = checker.getInvariants(methodName);
				int old_invariants = numberOfInvariants.get(methodName);
				if (invariants.size() < old_invariants) {
					System.out.println(methodName + " -> " + invariants.size()
							+ " invariants" + "("
							+ (old_invariants - invariants.size())
							+ " discarded)");
					numberOfInvariants.put(methodName, invariants.size());
				} else {
					System.out.println(methodName + " -> " + invariants.size()
							+ " invariants" + "(0 discarded)");
				}
			}
		}

		System.out.println("Writting new ginpink invariants file "
				+ checkedGinpinkInvFilename);
		checker.writeInvariants(checkedGinpinkInvFilename);

		if (discardedGinpinkInvFilename != null) {
			System.out.println("Writting discarded ginpink invariants file "
					+ discardedGinpinkInvFilename);
			checker.writeDiscardedInvariants(discardedGinpinkInvFilename);
		}

		System.out.println("Ginpink invariant checker done!");
	}
}
