package org.ginpink.eval;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;

import org.ginpink.GinPinkRuntime;
import org.ginpink.expr.JArrayAccess;
import org.ginpink.expr.JBinaryOperation;
import org.ginpink.expr.JBooleanConstant;
import org.ginpink.expr.JClassName;
import org.ginpink.expr.JConditionalExpr;
import org.ginpink.expr.JDot;
import org.ginpink.expr.JExpr;
import org.ginpink.expr.JExprVisitor;
import org.ginpink.expr.JIdentifier;
import org.ginpink.expr.JInstanceField;
import org.ginpink.expr.JInstanceMethodCall;
import org.ginpink.expr.JIntegerConstant;
import org.ginpink.expr.JLongConstant;
import org.ginpink.expr.JMethodCall;
import org.ginpink.expr.JNullExpr;
import org.ginpink.expr.JOperator;
import org.ginpink.expr.JStaticMethodCall;
import org.ginpink.expr.JStringLiteral;
import org.ginpink.expr.JUnaryOperation;
import org.ginpink.expr.JVar;
import org.ginpink.util.Reflection;

public final class ExprEvaluator implements JExprVisitor<Object> {

	private static final String LENGTH = "length";

	private static final Object NULL_VALUE = null;

	private final JavaHeap javaheap;

	public ExprEvaluator(JavaHeap javaheap) {
		this.javaheap = javaheap;
	}

	@Override
	public Object visit(JVar e) {
		Object value = javaheap.getObject(e.getVarName());
		return value;
	}

	@Override
	public Object visit(JBooleanConstant e) {
		return e.booleanValue();

	}

	@Override
	public Object visit(JIntegerConstant e) {
		return e.intValue();
	}

	@Override
	public Object visit(JLongConstant e) {
		return e.longValue();
	}

	@Override
	public Object visit(JInstanceMethodCall e) {
		String method_name = e.getMethodName();
		Object concrete_receiver = e.getReceiver().accept(this);
		LinkedList<JExpr> argumentExprList = e.getArguments();
		Object[] concreteArguments = getConcreteArguments(argumentExprList);
		Class<?>[] parameterTypes = Reflection.getParameterTypes(
				argumentExprList, concreteArguments);

		if (concrete_receiver == null) {
			throw new EvaluationException("receiver is null");
		}

		final Class<? extends Object> clazz = concrete_receiver.getClass();
		try {
			return Reflection.invokeMethod(clazz, method_name,
					concrete_receiver, concreteArguments, parameterTypes);
		} catch (InvocationTargetException ex) {
			throw new EvaluationException(ex);
		}
	}

	private Object[] getConcreteArguments(LinkedList<JExpr> argumentExprList) {
		Object[] concreteArguments = new Object[argumentExprList.size()];
		for (int i = 0; i < argumentExprList.size(); i++) {
			JExpr argumentExpr = argumentExprList.get(i);
			Object concrete_argument = argumentExpr.accept(this);
			concreteArguments[i] = concrete_argument;
		}
		return concreteArguments;
	}

	@Override
	public Object visit(JStaticMethodCall e) {
		String class_name = e.getClassName();
		String method_name = e.getMethodName();

		LinkedList<JExpr> argumentExprList = e.getArguments();
		Object[] concreteArguments = getConcreteArguments(argumentExprList);
		Class<?>[] parameterTypes = Reflection.getParameterTypes(
				argumentExprList, concreteArguments);

		ClassLoader class_loader = GinPinkRuntime.getClassLoader();
		try {
			Class<? extends Object> clazz = class_loader.loadClass(class_name);
			return Reflection.invokeMethod(clazz, method_name, null,
					concreteArguments, parameterTypes);
		} catch (ClassNotFoundException e1) {
			throw new RuntimeException(e1);
		} catch (InvocationTargetException ex) {
			throw new EvaluationException(ex);
		}

	}

	@Override
	public Object visit(JUnaryOperation e) {
		JOperator op = e.getOperator();
		JExpr expr = e.getExpr();
		Object concrete_expr = expr.accept(this);
		switch (op) {
		case LOGICAL_NOT: {
			Boolean concrete_boolean = (Boolean) concrete_expr;
			boolean ret_val = !concrete_boolean.booleanValue();
			return ret_val;
		}
		case UNARY_PLUS: {
			Integer concrete_integer = (Integer) concrete_expr;
			return concrete_integer;
		}
		case UNARY_MINUS: {
			Integer concrete_integer = (Integer) concrete_expr;
			int concrete_int_value = concrete_integer.intValue();
			int ret_val = -concrete_int_value;
			return ret_val;
		}
		default:
			throw new RuntimeException("Unhandled unary operator "
					+ op.toString());
		}
	}

	@Override
	public Object visit(JIdentifier e) {
		throw new IllegalArgumentException(
				"This class should not be executed using an object of type "
						+ e.getClass().getCanonicalName());
	}

	@Override
	public Object visit(JDot e) {
		throw new IllegalArgumentException(
				"This class should not be executed using an object of type "
						+ e.getClass().getCanonicalName());

	}

	@Override
	public Object visit(JMethodCall e) {
		throw new IllegalArgumentException(
				"This class should not be executed using an object of type "
						+ e.getClass().getCanonicalName());
	}

	@Override
	public Object visit(JConditionalExpr e) {
		Object predVal = e.getPredExpr().accept(this);
		boolean predBooleanVal = ((Boolean) predVal).booleanValue();
		Object ret_val;
		if (predBooleanVal == true) {
			ret_val = e.getTrueExpr().accept(this);
		} else {
			ret_val = e.getFalseExpr().accept(this);
		}
		return ret_val;
	}

	@Override
	public Object visit(JNullExpr e) {
		return NULL_VALUE;
	}

	@Override
	public Object visit(JArrayAccess e) {
		Object array_object = e.getArrayExpr().accept(this);
		Object index_object = e.getIndexExpr().accept(this);

		int index = (Integer) index_object;

		try {
			Object ret_val;
			if (e.getType().isInt()) {
				ret_val = Array.getInt(array_object, index);
			} else if (e.getType().isBoolean()) {
				ret_val = Array.getBoolean(array_object, index);
			} else if (!e.getType().isPrimitive()) {
				ret_val = Array.get(array_object, index);
			} else {
				throw new RuntimeException(
						"Array access not implemented yet for primitive type "
								+ e.getType());
			}
			return ret_val;
		} catch (ArrayIndexOutOfBoundsException ex) {
			throw new EvaluationException(ex);
		}
	}

	@Override
	public Object visit(JClassName e) {
		throw new RuntimeException("not implemented yet");
	}

	@Override
	public Object visit(JBinaryOperation e) {
		Object leftVal = e.getLeft().accept(this);
		Object rightVal = e.getRight().accept(this);

		switch (e.getOperator()) {
		case REM: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			int intValue = leftInteger % rightInteger;
			return (Integer) intValue;
		}
		case SUB: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			int intValue = leftInteger - rightInteger;
			return (Integer) intValue;
		}
		case ADD: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			int intValue = leftInteger + rightInteger;
			return (Integer) intValue;
		}
		case DIV: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			int intValue = leftInteger / rightInteger;
			return (Integer) intValue;
		}
		case MULT: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			int intValue = leftInteger * rightInteger;
			return (Integer) intValue;
		}
		case GREATER_THAN: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			if (leftInteger > rightInteger) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}
		}
		case GREATER_THAN_EQUAL: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			if (leftInteger >= rightInteger) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}
		}
		case LESS_THAN: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			if (leftInteger < rightInteger) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}
		}
		case LESS_THAN_EQUAL: {
			Integer leftInteger = (Integer) leftVal;
			Integer rightInteger = (Integer) rightVal;
			if (leftInteger <= rightInteger) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}
		}

		case EQUAL: {

			if (e.getLeft().getType().isLong()
					&& e.getRight().getType().isLong()) {
				long leftLong = (Long) leftVal;
				long rightLong = (Long) rightVal;
				if (leftLong == rightLong) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}
			} else if (e.getLeft().getType().isInt()
					&& e.getRight().getType().isInt()) {
				int leftInteger = (Integer) leftVal;
				int rightInteger = (Integer) rightVal;
				if (leftInteger == rightInteger) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}

			} else if (e.getLeft().getType().isBoolean()
					&& e.getRight().getType().isBoolean()) {
				boolean leftBoolean = (Boolean) leftVal;
				boolean rightBoolean = (Boolean) rightVal;
				if (leftBoolean == rightBoolean) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}

			} else {

				if (leftVal == rightVal) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}
			}
		}
		case NOT_EQUAL: {

			if (e.getLeft().getType().isLong()
					&& e.getRight().getType().isLong()) {
				long leftLong = (Long) leftVal;
				long rightLong = (Long) rightVal;
				if (leftLong != rightLong) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}
			} else if (e.getLeft().getType().isInt()
					&& e.getRight().getType().isInt()) {
				int leftInteger = (Integer) leftVal;
				int rightInteger = (Integer) rightVal;
				if (leftInteger != rightInteger) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}

			} else if (e.getLeft().getType().isBoolean()
					&& e.getRight().getType().isBoolean()) {
				boolean leftBoolean = (Boolean) leftVal;
				boolean rightBoolean = (Boolean) rightVal;
				if (leftBoolean != rightBoolean) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}

			} else {

				if (leftVal != rightVal) {
					return Boolean.TRUE;
				} else {
					return Boolean.FALSE;
				}
			}
		}
		case LOGICAL_AND: {
			boolean leftBoolean = (Boolean) leftVal;
			boolean rightBoolean = (Boolean) rightVal;
			if (leftBoolean && rightBoolean) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}
		}
		case LOGICAL_OR: {
			boolean leftBoolean = (Boolean) leftVal;
			boolean rightBoolean = (Boolean) rightVal;
			if (leftBoolean || rightBoolean) {
				return Boolean.TRUE;
			} else {
				return Boolean.FALSE;
			}

		}
		case UNARY_MINUS:
		case LOGICAL_NOT:
		case UNARY_PLUS: {
			throw new RuntimeException(
					"Unexpected unary operator in binary operation "
							+ e.getOperator());
		}

		}
		throw new RuntimeException("Unhandled binary operator "
				+ e.getOperator());
	}

	@Override
	public Object visit(JInstanceField e) {
		JExpr receiver = e.getInstanceExpr();
		Object concrete_instance = receiver.accept(this);

		if (concrete_instance == null) {
			throw new EvaluationException("The field expression for field "
					+ e.getFieldName() + " evalutated on null instance");
		}

		Class<? extends Object> concrete_class = concrete_instance.getClass();
		if (concrete_class.isArray()) {

			if (!e.getFieldName().equals(LENGTH)) {
				throw new RuntimeException("Unknown method for array class "
						+ e.getFieldName());
			}

			int array_length = Array.getLength(concrete_instance);
			return array_length;

		} else {

			try {

				Field concrete_field = concrete_class.getDeclaredField(e
						.getFieldName());
				boolean isAccessible = concrete_field.isAccessible();
				concrete_field.setAccessible(true);
				Object ret_val = concrete_field.get(concrete_instance);
				concrete_field.setAccessible(isAccessible);
				return ret_val;

			} catch (SecurityException e1) {
				throw new RuntimeException(e1);
			} catch (NoSuchFieldException e1) {
				throw new RuntimeException(e1);
			} catch (IllegalArgumentException e1) {
				throw new RuntimeException(e1);
			} catch (IllegalAccessException e1) {
				throw new RuntimeException(e1);
			}
		}

	}

	@Override
	public Object visit(JStringLiteral e) {
		return e.getStringLiteral();
	}

}
