package org.ginpink.eval;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Set;

import org.ginpink.GinPinkRuntime;
import org.ginpink.daikon.Snapshot;
import org.ginpink.daikon.SnapshotBooleanValue;
import org.ginpink.daikon.SnapshotIntValue;
import org.ginpink.daikon.SnapshotLongValue;
import org.ginpink.daikon.SnapshotNonNullReference;
import org.ginpink.daikon.SnapshotNullReference;
import org.ginpink.daikon.SnapshotValue;

public final class JavaHeapBuilder {

	public JavaHeap buildJavaHeap(Snapshot snapshot) {

		JavaHeap javaheap = new JavaHeap();

		Set<String> parameters = snapshot.parameters();
		for (String parameterName : parameters) {

			SnapshotValue snapshotValue = snapshot.getVariable(parameterName);

			if (snapshotValue instanceof SnapshotIntValue) {
				SnapshotIntValue snapshotIntValue = (SnapshotIntValue) snapshotValue;
				int intValue = snapshotIntValue.intValue();
				javaheap.setInt(parameterName, intValue);
			} else if (snapshotValue instanceof SnapshotLongValue) {
				SnapshotLongValue snapshotLongValue = (SnapshotLongValue) snapshotValue;
				long longValue = snapshotLongValue.longValue();
				javaheap.setLong(parameterName, longValue);

			} else if (snapshotValue instanceof SnapshotBooleanValue) {
				SnapshotBooleanValue snapshotIntValue = (SnapshotBooleanValue) snapshotValue;
				boolean intValue = snapshotIntValue.booleanValue();
				javaheap.setBoolean(parameterName, intValue);

			} else if (snapshotValue instanceof SnapshotNullReference) {
				javaheap.setNull(parameterName);

			} else if (snapshotValue instanceof SnapshotNonNullReference) {

				SnapshotNonNullReference ref = (SnapshotNonNullReference) snapshotValue;
				Object concreteValue = getOrCreateObject(ref);
				if (concreteValue instanceof int[]) {
					int[] intArray = (int[]) concreteValue;
					javaheap.setArrayInt(parameterName, intArray);
				} else if (concreteValue instanceof long[]) {
					long[] longArray = (long[]) concreteValue;
					javaheap.setArrayLong(parameterName, longArray);
				} else {
					javaheap.setObject(parameterName, concreteValue);
				}
			} else {
				throw new RuntimeException("Unhandled data type "
						+ snapshotValue.getClass().getName());
			}
		}
		return javaheap;
	}

	private void createInstance(SnapshotNonNullReference ref) {
		String className = ref.getCanonicalClassName();

		ClassLoader classLoader = GinPinkRuntime.getClassLoader();
		try {
			Class<?> clazz = classLoader.loadClass(className);
			Object object = clazz.newInstance();
			ref.setConcreteValue(object);

			Set<String> fieldNames = ref.fieldNames();
			for (String field_name : fieldNames) {
				SnapshotValue snap_value = ref.getFieldValue(field_name);
				Field field = clazz.getDeclaredField(field_name);
				boolean isAccessible = field.isAccessible();
				field.setAccessible(true);
				if (snap_value instanceof SnapshotBooleanValue) {
					SnapshotBooleanValue snap_boolean_value = (SnapshotBooleanValue) snap_value;
					boolean booleanValue = snap_boolean_value.booleanValue();
					field.setBoolean(object, booleanValue);
				} else if (snap_value instanceof SnapshotIntValue) {
					SnapshotIntValue snap_int_value = (SnapshotIntValue) snap_value;
					int intValue = snap_int_value.intValue();
					field.setInt(object, intValue);
				} else if (snap_value instanceof SnapshotNullReference) {
					field.set(object, null);
				} else {
					SnapshotNonNullReference field_ref = (SnapshotNonNullReference) snap_value;
					Object field_object = getOrCreateObject(field_ref);
					field.set(object, field_object);
				}
				field.setAccessible(isAccessible);
			}

		} catch (ClassNotFoundException e) {
			// This should not happen
			throw new RuntimeException("ClassName was not found:" + className,
					e);
		} catch (InstantiationException e) {
			throw new RuntimeException(
					"Constructor with no parameters should be found", e);
		} catch (IllegalAccessException e) {
			throw new RuntimeException(
					"Constructor should be publicly available", e);
		} catch (SecurityException e) {
			throw new RuntimeException(
					"Constructor with no parameters should be found", e);
		} catch (NoSuchFieldException e) {
			throw new RuntimeException(
					"Constructor with no parameters should be found", e);
		}
	}

	private Object getOrCreateObject(SnapshotNonNullReference ref) {

		if (ref.getConcreteValue() == null) {
			if (ref.isArray()) {
				// array
				if (ref.getCanonicalClassName().equals(
						int[].class.getCanonicalName())) {
					createIntArray(ref);
				} else if (ref.getCanonicalClassName().equals(
						long[].class.getCanonicalName())) {
					createLongArray(ref);
				} else {
					createObjectArray(ref);
				}
			} else {
				String class_name = ref.getCanonicalClassName();
				if (instanceConstructors.containsKey(class_name)) {
					InstanceConstructor instanceConstructor = instanceConstructors
							.get(class_name);
					instanceConstructor.createInstance(ref);
				} else {
					createInstance(ref);
				}
			}
		}
		Object concreteValue = ref.getConcreteValue();
		return concreteValue;
	}

	private void createIntArray(SnapshotNonNullReference ref) {
		int array_size = ref.getArraySize();
		int[] int_array = new int[array_size];
		ref.setConcreteValue(int_array);
		for (int i = 0; i < array_size; i++) {
			SnapshotValue snapshot_value = ref.getItem(i);
			SnapshotIntValue snapshot_int_value = (SnapshotIntValue) snapshot_value;
			int_array[i] = snapshot_int_value.intValue();
		}
	}

	private void createLongArray(SnapshotNonNullReference ref) {
		int array_size = ref.getArraySize();
		long[] long_array = new long[array_size];
		ref.setConcreteValue(long_array);
		for (int i = 0; i < array_size; i++) {
			SnapshotValue snapshot_value = ref.getItem(i);
			SnapshotLongValue snapshot_long_value = (SnapshotLongValue) snapshot_value;
			long_array[i] = snapshot_long_value.longValue();
		}
	}

	private void createObjectArray(SnapshotNonNullReference ref) {
		int array_size = ref.getArraySize();
		String array_base_class_name = ref.getCanonicalBaseClassName();

		ClassLoader class_loader = ClassLoader.getSystemClassLoader();
		Class<?> base_class;
		try {
			base_class = class_loader.loadClass(array_base_class_name);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException("Couldn't create array with base "
					+ array_base_class_name, e);
		}
		Object array_object = Array.newInstance(base_class, array_size);

		ref.setConcreteValue(array_object);
		for (int i = 0; i < array_size; i++) {
			SnapshotValue snapshot_value = ref.getItem(i);
			if (snapshot_value instanceof SnapshotNonNullReference) {
				SnapshotNonNullReference non_null_item = (SnapshotNonNullReference) snapshot_value;
				Object concrete_item = getOrCreateObject(non_null_item);
				Array.set(array_object, i, concrete_item);
			}

		}
	}

	private HashMap<String, InstanceConstructor> instanceConstructors = new HashMap<String, InstanceConstructor>();

	public void registerInstanceConstructor(
			InstanceConstructor instanceConstructor) {
		instanceConstructors.put(instanceConstructor.getClassName(),
				instanceConstructor);
	}

	public void unregisterInstanceConstructor(
			InstanceConstructor instanceConstructor) {
		instanceConstructors.remove(instanceConstructor.getClassName());
	}
}
