package org.ginpink.parser;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.TreeVisitorAction;
import org.ginpink.expr.JArrayAccess;
import org.ginpink.expr.JBinaryOperation;
import org.ginpink.expr.JBooleanConstant;
import org.ginpink.expr.JConditionalExpr;
import org.ginpink.expr.JDot;
import org.ginpink.expr.JExpr;
import org.ginpink.expr.JIdentifier;
import org.ginpink.expr.JIntegerConstant;
import org.ginpink.expr.JMethodCall;
import org.ginpink.expr.JNullExpr;
import org.ginpink.expr.JOperator;
import org.ginpink.expr.JStringLiteral;
import org.ginpink.expr.JUnaryOperation;

public final class ExprVisitorAction implements TreeVisitorAction {

	private Stack<List<JExpr>> stack = new Stack<List<JExpr>>();

	@Override
	public Object pre(Object t) {
		return t;
	}

	@Override
	public Object post(Object t) {
		if (t == null)
			return null;

		CommonTree node = (CommonTree) t;
		switch (node.getType()) {
		case JavaParser.EXPR: {
			// do nothing
			return t;
		}
		case JavaParser.PARENTESIZED_EXPR: {
			// do nothing
			return t;
		}
		case JavaParser.DECIMAL_LITERAL: {
			String integer_str = node.getText();
			Integer integer = Integer.parseInt(integer_str);
			int intValue = integer.intValue();
			JIntegerConstant expr = new JIntegerConstant(intValue);
			stack.push(Collections.<JExpr> singletonList(expr));
			return t;
		}
		case JavaParser.UNARY_MINUS: {
			JExpr expr = stack.pop().get(0);
			JUnaryOperation negated_int_constant = new JUnaryOperation(
					JOperator.UNARY_MINUS, expr);
			stack.push(Collections.<JExpr> singletonList(negated_int_constant));
			return t;
		}
		case JavaParser.TRUE: {
			JBooleanConstant true_constant = new JBooleanConstant(true);
			stack.push(Collections.<JExpr> singletonList(true_constant));
			return t;
		}
		case JavaParser.FALSE: {
			JBooleanConstant false_constant = new JBooleanConstant(false);
			stack.push(Collections.<JExpr> singletonList(false_constant));
			return t;
		}
		case JavaParser.LOGICAL_NOT: {
			JExpr expr = stack.pop().get(0);
			JUnaryOperation logical_not = new JUnaryOperation(
					JOperator.LOGICAL_NOT, expr);
			stack.push(Collections.<JExpr> singletonList(logical_not));
			return t;
		}
		case JavaParser.PLUS: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);
			JBinaryOperation add = new JBinaryOperation(left, JOperator.ADD,
					right);
			stack.push(Collections.<JExpr> singletonList(add));
			return t;
		}
		case JavaParser.MINUS: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);
			JBinaryOperation add = new JBinaryOperation(left, JOperator.SUB,
					right);
			stack.push(Collections.<JExpr> singletonList(add));
			return t;
		}
		case JavaParser.STAR: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);
			JBinaryOperation add = new JBinaryOperation(left, JOperator.MULT,
					right);
			stack.push(Collections.<JExpr> singletonList(add));
			return t;
		}
		case JavaParser.DIV: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);
			JBinaryOperation add = new JBinaryOperation(left, JOperator.DIV,
					right);
			stack.push(Collections.<JExpr> singletonList(add));
			return t;
		}
		case JavaParser.MOD: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);
			JBinaryOperation add = new JBinaryOperation(left, JOperator.REM,
					right);
			stack.push(Collections.<JExpr> singletonList(add));
			return t;
		}
		case JavaParser.IDENT: {
			String ident_str = node.getText();
			JIdentifier ident = new JIdentifier(ident_str);
			stack.push(Collections.<JExpr> singletonList(ident));
			return t;
		}
		case JavaParser.DOT: {
			JIdentifier rightExpr = (JIdentifier) stack.pop().get(0);
			JExpr leftExpr = stack.pop().get(0);

			JDot fieldExpr = new JDot(leftExpr, rightExpr);
			stack.push(Collections.<JExpr> singletonList(fieldExpr));
			return t;
		}
		case JavaParser.ARGUMENT_LIST: {
			int numberOfChildren = node.getChildCount();
			LinkedList<JExpr> arguments = new LinkedList<JExpr>();
			for (int i = 0; i < numberOfChildren; i++) {
				JExpr argument = stack.pop().get(0);
				arguments.addFirst(argument);
			}
			stack.push(arguments);
			return t;
		}
		case JavaParser.METHOD_CALL: {
			List<JExpr> arguments = stack.pop();
			JExpr expr = stack.pop().get(0);
			final JMethodCall method_call = new JMethodCall(expr, arguments);
			stack.push(Collections.<JExpr> singletonList(method_call));
			return t;
		}
		case JavaParser.QUESTION: {
			JExpr falseExpr = stack.pop().get(0);
			JExpr trueExpr = stack.pop().get(0);
			JExpr predExpr = stack.pop().get(0);

			JConditionalExpr cond_expr = new JConditionalExpr(predExpr,
					trueExpr, falseExpr);
			stack.push(Collections.<JExpr> singletonList(cond_expr));
			return t;
		}
		case JavaParser.NULL: {
			stack.push(Collections.<JExpr> singletonList(JNullExpr
					.getNullExpr()));
			return t;
		}
		case JavaParser.EQUAL: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.EQUAL, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.LESS_THAN: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.LESS_THAN, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.GREATER_THAN: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.GREATER_THAN, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.GREATER_OR_EQUAL: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.GREATER_THAN_EQUAL, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.LESS_OR_EQUAL: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.LESS_THAN_EQUAL, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.NOT_EQUAL: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation equal_op = new JBinaryOperation(left,
					JOperator.NOT_EQUAL, right);

			stack.push(Collections.<JExpr> singletonList(equal_op));
			return t;
		}
		case JavaParser.ARRAY_ELEMENT_ACCESS: {
			JExpr indexExpr = stack.pop().get(0);
			JExpr arrayExpr = stack.pop().get(0);

			JArrayAccess array_access = new JArrayAccess(arrayExpr, indexExpr);

			stack.push(Collections.<JExpr> singletonList(array_access));
			return t;
		}
		case JavaParser.STRING_LITERAL: {
			String string = node.getText();
			String noquotes = string.substring(1, string.length() - 1);
			JStringLiteral stringLiteral = new JStringLiteral(noquotes);

			stack.push(Collections.<JExpr> singletonList(stringLiteral));
			return t;
		}
		case JavaParser.LOGICAL_AND: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation logical_and = new JBinaryOperation(left,
					JOperator.LOGICAL_AND, right);
			stack.push(Collections.<JExpr> singletonList(logical_and));
			return t;
		}
		case JavaParser.LOGICAL_OR: {
			JExpr right = stack.pop().get(0);
			JExpr left = stack.pop().get(0);

			JBinaryOperation logical_and = new JBinaryOperation(left,
					JOperator.LOGICAL_OR, right);
			stack.push(Collections.<JExpr> singletonList(logical_and));
			return t;
		}

		}
		throw new RuntimeException("unhandled expression type: "
				+ JavaParser.tokenNames[node.getType()]);
	}

	public JExpr getExpression() {
		if (stack.size() != 1 || stack.peek().size() != 1)
			throw new IllegalStateException(
					"The building process is unfinished!");

		return stack.peek().get(0);
	}

}
