package org.ginpink.parser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeAdaptor;
import org.antlr.runtime.tree.TreeVisitor;
import org.ginpink.GinPinkRuntime;
import org.ginpink.Invariant;
import org.ginpink.LoopInvariantMethod;
import org.ginpink.expr.JExpr;
import org.ginpink.parser.JavaParser.expression_return;
import org.ginpink.typecheck.ExprTypeChecker;

public final class InvariantFileReader {

	private static final String EXEC_JML = "<--exec|JML-->";
	private static final String SEPARATOR_LINE = "===========================================================================";

	public InvariantFileReader() {
		this(null, null);
	}

	private final String className;
	private final String methodName;

	public InvariantFileReader(String className, String methodName) {
		this.className = className;
		this.methodName = methodName;
	}

	private HashMap<String, List<JExpr>> parsed_expressions = new HashMap<String, List<JExpr>>();

	private HashMap<String, Integer> ignored_expressions = new HashMap<String, Integer>();

	public List<LoopInvariantMethod> readFile(String invFilename)
			throws IOException, RecognitionException {
		BufferedReader in = new BufferedReader(new FileReader(invFilename));
		LoopInvariantMethod current_loop_method = null;
		ExprTypeChecker current_type_checker = null;

		List<LoopInvariantMethod> loop_invariants = new LinkedList<LoopInvariantMethod>();

		while (in.ready()) {

			String str = in.readLine();
			str = str.trim();
			if (str.equals(SEPARATOR_LINE)) {
				if (!in.ready())
					break;

				String loopName = in.readLine();
				String className_dot_methodName = loopName.split("\\(")[0];
				String[] className_dot_methodName_parts = className_dot_methodName
						.split("\\.");

				String loopMethodName = className_dot_methodName_parts[className_dot_methodName_parts.length - 1];

				try {
					Method loopMethod = getMethod(className, loopMethodName);
					current_type_checker = new ExprTypeChecker(loopMethod);
					current_loop_method = new LoopInvariantMethod(loopMethod);
					loop_invariants.add(current_loop_method);
				} catch (ClassNotFoundException e) {
					in.close();
					throw new RuntimeException(e);
				} catch (NoSuchMethodException e) {
					in.close();
					throw new RuntimeException(e);
				}

			} else {

				String java_string;
				String jml_string;
				
				
				if (str.contains(EXEC_JML)) {
					java_string = str.split(EXEC_JML.replace("|", "\\|"))[0];
					jml_string = str.split(EXEC_JML.replace("|", "\\|"))[1];
				} else {
					java_string = str;
					jml_string = str;
				}

				JExpr parsed_java_expr = parseExpr(java_string);
				JExpr type_checked_java_expr = parsed_java_expr
						.accept(current_type_checker);

				if (!type_checked_java_expr.isTypeChecked()) {
					in.close();
					throw new RuntimeException(
							"Unexpected: invariant does not type-check!");
				}

				Invariant invariant = new Invariant(type_checked_java_expr,
						java_string, jml_string);
				current_loop_method.addInvariant(invariant);
			}
		}
		in.close();
		return loop_invariants;
	}

	private Method getMethod(String className, String methodName)
			throws ClassNotFoundException, NoSuchMethodException {
		Class<?> clazz = GinPinkRuntime.getClassLoader().loadClass(className);
		Method[] methods = clazz.getDeclaredMethods();
		for (int i = 0; i < methods.length; i++) {
			Method method = methods[i];
			if (method.getName().equals(methodName)) {
				return method;
			}
		}

		throw new NoSuchMethodException();
	}

	private JExpr parseExpr(String javaString) throws RecognitionException {
		ANTLRStringStream charStream = new ANTLRStringStream(javaString);
		JavaLexer javalexer = new JavaLexer(charStream);
		CommonTokenStream tokens = new CommonTokenStream();
		tokens.setTokenSource(javalexer);
		JavaParser javaparser = new JavaParser(tokens);

		expression_return expr_return = javaparser.expression();

		CommonTree common_tree = expr_return.getTree();

		TreeVisitor v = new TreeVisitor(new CommonTreeAdaptor());
		ExprVisitorAction actions = new ExprVisitorAction();
		v.visit(common_tree, actions);
		return actions.getExpression();
	}

	public Set<String> methodSet() {
		return this.parsed_expressions.keySet();
	}

	public List<JExpr> getInvariants(String methodName) {
		return this.parsed_expressions.get(methodName);
	}

	public int getIgnoredExpressionsCount(String methodName) {
		return this.ignored_expressions.get(methodName);
	}
}
