package org.ginpink.typecheck;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.ginpink.GinPinkRuntime;
import org.ginpink.expr.JArrayAccess;
import org.ginpink.expr.JBinaryOperation;
import org.ginpink.expr.JBooleanConstant;
import org.ginpink.expr.JClassName;
import org.ginpink.expr.JConditionalExpr;
import org.ginpink.expr.JDot;
import org.ginpink.expr.JExpr;
import org.ginpink.expr.JExprVisitor;
import org.ginpink.expr.JIdentifier;
import org.ginpink.expr.JInstanceField;
import org.ginpink.expr.JInstanceMethodCall;
import org.ginpink.expr.JIntegerConstant;
import org.ginpink.expr.JLongConstant;
import org.ginpink.expr.JMethodCall;
import org.ginpink.expr.JNullExpr;
import org.ginpink.expr.JStaticMethodCall;
import org.ginpink.expr.JStringLiteral;
import org.ginpink.expr.JUnaryOperation;
import org.ginpink.expr.JVar;
import org.ginpink.util.Reflection;

public final class ExprVisibilityChecker implements JExprVisitor<Boolean> {

	private static final String LENGTH = "length";

	public ExprVisibilityChecker(Method method) {
		this.reflect_method = method;
	}

	@Override
	public Boolean visit(JBinaryOperation e) {
		boolean ret_val_left = e.getLeft().accept(this);
		boolean ret_val_right = e.getRight().accept(this);
		return ret_val_left && ret_val_right;
	}

	@Override
	public Boolean visit(JIntegerConstant e) {
		return true;
	}

	@Override
	public Boolean visit(JLongConstant e) {
		return true;
	}

	@Override
	public Boolean visit(JVar e) {
		return true;
	}

	@Override
	public Boolean visit(JInstanceMethodCall e) {
		boolean visibility_check = e.getReceiver().accept(this);
		if (!visibility_check) {
			return false;
		}
		for (JExpr expr : e.getArguments()) {
			visibility_check = expr.accept(this);
			if (!visibility_check) {
				return false;
			}
		}

		Class<?> declaring_method_class = reflect_method.getDeclaringClass();

		Class<?> receiver_class = e.getReceiver().getType().getJavaClass();
		String methodName = e.getMethodName();
		Class<?>[] arg_types = new Class<?>[e.getArguments().size()];
		for (int i = 0; i < e.getArguments().size(); i++) {
			JExpr expr = e.getArguments().get(i);
			Class<?> java_class = expr.getType().getJavaClass();
			arg_types[i] = java_class;
		}

		try {
			Method instance_method = Reflection.getDeclaredMethod(
					receiver_class, methodName, arg_types);
			Class<?> access_instance_class = instance_method
					.getDeclaringClass();
			int access_modifiers = instance_method.getModifiers();
			boolean is_visible = isVisible(declaring_method_class,
					access_instance_class, access_modifiers);
			return is_visible;

		} catch (NoSuchMethodException e1) {
			throw new RuntimeException(e1);
		}
	}

	@Override
	public Boolean visit(JUnaryOperation e) {
		boolean ret_val = e.getExpr().accept(this);
		return ret_val;
	}

	@Override
	public Boolean visit(JBooleanConstant e) {
		return true;
	}

	@Override
	public Boolean visit(JStaticMethodCall e) {
		for (JExpr expr : e.getArguments()) {
			boolean visibility_check = expr.accept(this);
			if (!visibility_check) {
				return false;
			}
		}

		Class<?> declaring_method_class = reflect_method.getDeclaringClass();

		String class_name = e.getClassName();
		String method_name = e.getMethodName();
		Class<?>[] arg_types = new Class<?>[e.getArguments().size()];
		for (int i = 0; i < e.getArguments().size(); i++) {
			JExpr expr = e.getArguments().get(i);
			arg_types[i] = expr.getType().getJavaClass();
		}
		ClassLoader class_loader = GinPinkRuntime.getClassLoader();
		try {
			Class<? extends Object> access_instance_class = class_loader
					.loadClass(class_name);
			Method static_method = Reflection.getDeclaredMethod(
					access_instance_class, method_name, arg_types);
			int access_modifiers = static_method.getModifiers();
			boolean is_visible = isVisible(declaring_method_class,
					access_instance_class, access_modifiers);
			return is_visible;
		} catch (ClassNotFoundException e1) {
			throw new RuntimeException(e1);
		} catch (SecurityException e1) {
			throw new RuntimeException(e1);
		} catch (NoSuchMethodException e1) {
			throw new RuntimeException(e1);
		}
	}

	private final Method reflect_method;

	@Override
	public Boolean visit(JInstanceField e) {
		JExpr instanceExpr = e.getInstanceExpr();
		Class<?> instance_class = instanceExpr.getType().getJavaClass();
		Class<?> declaring_method_class = reflect_method.getDeclaringClass();

		if (reflect_method.getDeclaringClass().equals(instance_class)) {
			return true;
		}
		String fieldName = e.getFieldName();
		if (fieldName.equals(LENGTH)) {
			return true;
		}

		try {
			Field field = instance_class.getDeclaredField(fieldName);
			int modifiers = field.getModifiers();
			boolean is_visible = isVisible(declaring_method_class,
					instance_class, modifiers);
			return is_visible;
		} catch (SecurityException e1) {
			throw new RuntimeException(e1);
		} catch (NoSuchFieldException e1) {
			throw new RuntimeException(e1);
		}
	}

	private static boolean isVisible(Class<?> declaring_method_class,
			Class<?> access_instance_class, int access_modifiers) {
		if (Modifier.isPublic(access_modifiers)) {
			return true;
		} else if (Modifier.isPrivate(access_modifiers)) {
			boolean same_class = declaring_method_class
					.equals(access_instance_class);
			return same_class;
		} else if (Modifier.isProtected(access_modifiers)) {
			boolean isAssignableFrom = access_instance_class
					.isAssignableFrom(declaring_method_class);
			return isAssignableFrom;
		} else {
			//Default visbility 
			String instancePackageName = access_instance_class.getPackage()
					.getName();
			String methodPackageName = declaring_method_class.getPackage()
					.getName();
			boolean same_package = instancePackageName
					.equals(methodPackageName);
			return same_package;
		}
	}

	@Override
	public Boolean visit(JConditionalExpr e) {
		boolean pred_access = e.getPredExpr().accept(this);
		boolean true_branch_access = e.getTrueExpr().accept(this);
		boolean false_branch_access = e.getFalseExpr().accept(this);
		return pred_access && true_branch_access && false_branch_access;
	}

	@Override
	public Boolean visit(JNullExpr jNullExpr) {
		return true;
	}

	@Override
	public Boolean visit(JArrayAccess e) {
		return true;
	}

	@Override
	public Boolean visit(JIdentifier e) {
		throw new IllegalArgumentException(
				"The Java expression must be typechecked in order to check visibility");
	}

	@Override
	public Boolean visit(JDot e) {
		throw new IllegalArgumentException(
				"The Java expression must be typechecked in order to check visibility");
	}

	@Override
	public Boolean visit(JMethodCall e) {
		throw new IllegalArgumentException(
				"The Java expression must be typechecked in order to check visibility");
	}

	@Override
	public Boolean visit(JClassName e) {
		return true;
	}

	@Override
	public Boolean visit(JStringLiteral e) {
		return true;
	}

}
